#include "codegen.cpp"
#include "tokenizer.cpp"
#include "llvm/ADT/Optional.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/Module.h"
#include <iostream>

#define NEXT_TOK                                         \
    {                                                    \
        int ret = next_token(&tok);                      \
        if (ret != LEX_SUCCESS) {                        \
            printf("WARN: Unable to get next token\n"); \
            return nullptr;                              \
        }                                                \
    }

static std::vector<std::string> symbol_table;

/*
Basically the tokenizer function "get_token()" but with some evaluation of the
result.
*/
static int next_token(token_t* tok)
{
    int ret = get_token(tok);

    if (ret == LEX_SUCCESS) {
        printf("Token[type=%d, data=\"%s\"] at %d:%d\n", tok->tok_type, tok->tok_data.c_str(), tok->tok_line, tok->tok_col);
    }

    switch (ret) {
    case LEX_END:
        printf("Reached EOF\n");
        break;
    case LEX_UNKNOWN:
        printf("ERROR: Unknown token start found\n");
        break;
    }

    return ret;
}

/*
Parses the '(' letter.
*/
static int parse_bracket_round_open(token_t* tok)
{
    if (tok->tok_type != TOK_BRACKET_ROUND_OPEN) {
        printf("ERROR: Expected '(' at %d:%d\n", tok->tok_line, tok->tok_col);
        return 1;
    }

    return 0;
}

/*
Parses the ')' letter.
*/
static int parse_bracket_round_close(token_t* tok)
{
    if (tok->tok_type != TOK_BRACKET_ROUND_CLOSE) {
        printf("ERROR: Expected ')' at %d:%d\n", tok->tok_line, tok->tok_col);
        return 1;
    }

    return 0;
}

/*
Parsing a number literal which can be any valid c++ number string parsable by
"std::stoi()"

number ::= <c++ integer value>
*/
static std::unique_ptr<ASTIntegerExpression> parse_int_constant(token_t* tok)
{
    if (tok->tok_type != TOK_NUMBER) {
        printf("ERROR: Expected number at %d:%d\n", tok->tok_line, tok->tok_col);
        return nullptr;
    }

    int value = std::stoi(tok->tok_data);
    return llvm::make_unique<ASTIntegerExpression>(value);
}

static std::unique_ptr<ASTVariableExpression> parse_variable(token_t* tok)
{
    if (tok->tok_type != TOK_IDENT) {
        printf("ERROR: Expected identifier at %d:%d", tok->tok_line, tok->tok_col);
        return nullptr;
    }

    return llvm::make_unique<ASTVariableExpression>(tok->tok_data);
}

/*
return-statement ::= return(<number>) | return(<identifier>)
*/
static std::unique_ptr<ASTReturnExpression> parse_return_expression()
{
    token_t tok;

    NEXT_TOK;
    if (parse_bracket_round_open(&tok)) {
        return nullptr;
    }

    NEXT_TOK;
    std::unique_ptr<ASTExpression> expr;
    if (tok.tok_type == TOK_NUMBER) {
        expr = parse_int_constant(&tok);
        if (expr == nullptr) {
            return nullptr;
        }
    } else if (tok.tok_type == TOK_IDENT) {
        expr = parse_variable(&tok);
        if (expr == nullptr) {
            return nullptr;
        }
    } else {
        printf("ERROR: Expected number or identifier at %d:%d\n", tok.tok_line, tok.tok_col);
        return nullptr;
    }

    NEXT_TOK;
    if (parse_bracket_round_close(&tok)) {
        return nullptr;
    }

    return llvm::make_unique<ASTReturnExpression>(std::move(expr));
}

static int parse_identifier(token_t* tok, std::string* name)
{
    if (tok->tok_type != TOK_IDENT) {
        printf("ERROR: Expected identifier at %d:%d\n", tok->tok_line, tok->tok_col);
        return 1;
    }

    *name = tok->tok_data;
    return 0;
}

static std::unique_ptr<ASTVariableDeclarationExpression> parse_variable_declaration()
{
    token_t tok;

    NEXT_TOK;
    if (parse_bracket_round_open(&tok)) {
        return nullptr;
    }

    NEXT_TOK;
    std::string identifier;
    int err = parse_identifier(&tok, &identifier);
    if (err != 0) {
        return nullptr;
    }
    NEXT_TOK;
    std::unique_ptr<ASTIntegerExpression> expr = parse_int_constant(&tok);
    if (expr == nullptr) {
        return nullptr;
    }

    NEXT_TOK;
    if (parse_bracket_round_close(&tok)) {
        return nullptr;
    }

    symbol_table.push_back(identifier);

    return llvm::make_unique<ASTVariableDeclarationExpression>(identifier, std::move(expr));
}

void parse()
{
    // Make the llvm_module, which holds all the code.
    llvm_module = llvm::make_unique<llvm::Module>("small llvm compiler", llvm_context);

    llvm::FunctionType* FT = llvm::FunctionType::get(llvm::Type::getInt32Ty(llvm_context), false);

    llvm::Function* F = llvm::Function::Create(FT, llvm::Function::ExternalLinkage, "main", llvm_module.get());

    llvm::BasicBlock* llvm_start_point = llvm::BasicBlock::Create(llvm_context, F->getName(), F);
    llvm_builder.SetInsertPoint(llvm_start_point);

    int err = 0;
    do {
        token_t tok;
        int ret = next_token(&tok);

        if (ret == LEX_END) {
            break;
        } else if (ret == LEX_UNKNOWN) {
            printf("ERROR: Unknown token or token-start found\n");
            return;
        }

        switch (tok.tok_type) {
        case TOK_RET: {
            std::unique_ptr<ASTReturnExpression> expr = parse_return_expression();

            if (expr == nullptr) {
                printf("ERROR: Parsing return expression at %d:%d returned null-pointer\n", tok.tok_line, tok.tok_col);
                return;
            }

            expr->codegen();

            break;
        }
        case TOK_VAR: {
            std::unique_ptr<ASTVariableDeclarationExpression> expr = parse_variable_declaration();

            if (expr == nullptr) {
                printf("ERROR: Parsing variable declaration at %d:%d returned null-pointer\n", tok.tok_line, tok.tok_col);
                return;
            }

            expr->codegen();

            break;
        }
        default:
            printf("ERROR: Unsupported token type '%d' found\n", tok.tok_type);
            return;
        }
    } while (err == 0);

    printf("Finished parsing\n\n");

    write_llvm_file();
    write_obj_file();

    printf("\n");
}
